(* $Id: cmd.ml 680 2007-02-25 20:35:07Z gerd $
 * ----------------------------------------------------------------------
 * Copyright 2003 by Gerd Stolpmann.
 *
 * This file is part of godi-tools.
 *
 * godi-tools is free software; you can redistribute it and/or modify it
 * under the conditions in the file LICENSE.
 *
 * godi-tools is distributed WITHOUT ANY WARRANTY, as detailed in the
 * file LICENSE.
 *)

open Godi_types
open Godi_query
open Printf

exception Exit_code of int



let usage() =
  prerr_endline ("usage: godi_console [<command> [<options>]]\n" ^
		 "Without command: start interactive console\n" ^
		 "Available commands:\n" ^
		 "  update - Update the build instructions\n" ^
		 "  wish - Manipulate the list of user's wishes\n" ^
		 "  perform - Install or remove packages\n" ^
		 "  list - List packages by criterion\n" ^
                 "  dep - Check dependency on installed package\n" ^ 
		 "  delete - Remove packages (godi_delete backend)\n" ^
		 "  add - Install binary packages (godi_add backend)\n" ^ 
		 "  ptool - Process packages with the package tool\n" ^ 
		 "  make - Invoke the built-in 'make' utility (godi_make backend)\n" ^ 
		 "The commands have their own help texts, use -help\n" ^
		 "to see them, e.g.\n" ^
		 "  godi_console perform -help");
  exit 1
;;

let opt_re = Str.regexp "^\\([^=:]+\\):\\([^=]+\\)=\\(.*\\)$";;
let ropt_re = Str.regexp "^\\([^=:]+\\):\\([^=]+\\)$";;

let split_re = Str.regexp "[ \t\r\n]+";;
let split s = Str.split split_re s;;

let word_by_word f s =
  let l = split s in
  List.iter f l
;;


let inst_pkg_by_name name =
  let (n, version, rev) =
    Godi_versions.split_name_version name in
  let pkg =
    try
      get_this_installed_package n
    with
      | Not_found ->
	  user_error ("No such installed package: " ^ n) in
  if version <> "" then (
    let dep = Godi_versions.parse_dependency (n ^ "==" ^ version) in
    if not (Godi_query.matches_dep pkg dep) then
      user_error ("Package not installed in requested version: " ^ name);
    (* We do not check the revision number... *)
  );
  pkg
;;


let cmd_update() =
  Arg.current := 1;
  Arg.parse 
      []
      (fun _ -> raise (Arg.Bad "Unexpected argument"))
      "usage: godi_console update [<options>]";
  Godi_update.update_available();
  Godi_update.update_build_instructions()
;;


let cmd_wish() =
  let wishes = ref (Godi_wishes.read_wishes ~pref_user_error:true ()) in
  let modified = ref false in

  let check_src_pkg pkgname =
    let src_pkgs = scan_source_packages() in
    if not (List.exists (fun p -> pkg_name p = pkgname) src_pkgs) then
      user_error ("No such source package: " ^ pkgname)
  in

  let add_build_wish pkgname =
    check_src_pkg pkgname;
    let wish = `Build pkgname in
    if not (List.mem wish !wishes) then (
      wishes := !wishes @ [ wish ];
      modified := true
    )
  in
 
  let build_all_installed() =
    let src_pkgs = scan_source_packages() in
    let src_pkgs_map = PkgMap.make (List.map (fun pkg -> (pkg,())) src_pkgs) in
    let inst_pkgs = get_installed_packages() in
    List.iter
      (fun pkg ->
	 let n = pkg_name pkg in
	 let spkgs = PkgMap.find_by ~name:n src_pkgs_map in
	 if spkgs <> [] then add_build_wish n
      )
      inst_pkgs
  in

  let rem_build_wish pkgname =
    let wish = `Build pkgname in
    wishes := List.filter (fun w -> w <> wish) !wishes;
    modified := true
  in
    
  let add_remove_wish pkgname =
    let pkg = inst_pkg_by_name pkgname in
    let wish = `Remove (pkg_name pkg) in
    if not (List.mem wish !wishes) then (
      wishes := !wishes @ [ wish ];
      modified := true
    );
  in
 
  let rem_remove_wish pkgname =
    let (name, _, _) = Godi_versions.split_name_version pkgname in
    let wish = `Remove name in
    wishes := List.filter (fun w -> w <> wish) !wishes;
    modified := true
  in
    
  let add_option opt =
    if Str.string_match opt_re opt 0 then (
      let pkgname = Str.matched_group 1 opt in
      let optname = Str.matched_group 2 opt in
      let optvalue = Str.matched_group 3 opt in
      check_src_pkg pkgname;
      let wl = List.filter (fun w -> 
			      match w with
				  `Option(p,n,_) -> p <> pkgname || n <> optname
				| _ -> true) !wishes in
      wishes := wl @ [ `Option(pkgname,optname,optvalue) ];
      modified := true
    )
    else
      user_error ("Syntax error in -option " ^ opt)
  in
    
  let rem_option opt =
    if Str.string_match ropt_re opt 0 then (
      let pkgname = Str.matched_group 1 opt in
      let optname = Str.matched_group 2 opt in
      let wl = List.filter (fun w -> 
			      match w with
				  `Option(p,n,_) -> p <> pkgname || n <> optname
				| _ -> true) !wishes in
      wishes := wl;
      modified := true
    )
    else
      user_error ("Syntax error in -no-option " ^ opt)
  in

  let reset() =
    wishes := [];
    modified := true
  in

  let print() =
    (* Sort wishes: *)
    let pkg_of_wish =
      function
	  `Build p -> p
	| `Remove p -> p
	| `Option(p,_,_) -> p in
    let cat_of_wish =
      function
	  `Build _ -> 1
	| `Remove _ -> 0
	| `Option(_,_,_) -> 2 in
    let optname_of_wish =
      function
	  `Build _ -> ""
	| `Remove _ -> ""
	| `Option(n,_,_) -> n in
    let cmp w1 w2 =
      let p1 = pkg_of_wish w1 in
      let p2 = pkg_of_wish w2 in
      match compare p1 p2 with
	  0 ->
	    ( match cat_of_wish w1 - cat_of_wish w2 with
		  0 ->
		    compare (optname_of_wish w1) (optname_of_wish w2)
		| n -> n
	    )
	| n -> n
    in
    let wl = List.sort cmp !wishes in
    (* Finally print wishes: *)
    List.iter
      (function
	  `Build p ->
	    printf "-build  %s\n" p
	| `Remove p ->
	    printf "-remove %s\n" p
	| `Option(p,n,v) ->
	    printf "-option %s:%s=%s\n" p n v
      )
      wl
  in

  Arg.current := 1;
  Arg.parse 
      [ "-build", Arg.String (word_by_word add_build_wish),
	       "<pkg>                 Add wish to build <pkg>";
	"-no-build", Arg.String (word_by_word rem_build_wish),
	          "<pkg>              Withdraw wish to build <pkg>";
	"-remove", Arg.String (word_by_word add_remove_wish),
	        "<pkg>                Add wish to remove <pkg>";
	"-no-remove", Arg.String (word_by_word rem_remove_wish),
	           "<pkg>             Withdraw wish to remove <pkg>";
	"-option", Arg.String add_option,
	        "<pkg>:<name>=<val>   Add option <name>=<val> for <pkg>";
	"-no-option", Arg.String rem_option,
	           "<pkg>:<name>      Withdraw option <name> for <pkg>";
	"-rebuild", Arg.Unit build_all_installed,
	         "                    Wish to rebuild all installed packages";
	"-reset", Arg.Unit reset,
	       "                      Withdraw all wishes";
	"-print", Arg.Unit print,
	       "                      Print resulting wish list";
      ]
      (fun _ -> raise (Arg.Bad "Unexpected argument"))
      "usage: godi_console wish <options>";

  if !modified then
    Godi_wishes.write_wishes ~pref_user_error:true !wishes
;;


let cmd_perform() =
  let wishes = ref [] in
  let use_wishlist = ref false in
  let dry_run = ref false in
  let build_all_installed = ref false in
  let newer = ref false in
  let binary_additions = ref [] in
  let disable_builder = ref false in
  let replace = ref `Never in

  let check_src_pkg pkgname =
    let src_pkgs = scan_source_packages() in
    if not (List.exists (fun p -> pkg_name p = pkgname) src_pkgs) then
      user_error ("No such source package: " ^ pkgname)
  in

  let add_build_wish pkgname =
    check_src_pkg pkgname;
    let wish = `Build pkgname in
    if not (List.mem wish !wishes) then (
      wishes := !wishes @ [ wish ];
    )
  in
 
  let add_remove_wish pkgname =
    let pkg = inst_pkg_by_name pkgname in
    let wish = `Remove (pkg_name pkg) in
    if not (List.mem wish !wishes) then (
      wishes := !wishes @ [ wish ];
    );
  in
 
  let add_option opt =
    if Str.string_match opt_re opt 0 then (
      let pkgname = Str.matched_group 1 opt in
      let optname = Str.matched_group 2 opt in
      let optvalue = Str.matched_group 3 opt in
      check_src_pkg pkgname;
      let wl = List.filter (fun w -> 
			      match w with
				  `Option(p,n,_) -> p <> pkgname || n <> optname
				| _ -> true) !wishes in
      wishes := wl @ [ `Option(pkgname,optname,optvalue) ];
    )
    else
      user_error ("Syntax error in -option " ^ opt)
  in

  let add_wishes() =
    use_wishlist := true
  in

  let add_package file =
    if Sys.file_exists file then (
      binary_additions := !binary_additions @ [file]
    )
    else
      user_error ("File does not exist: " ^ file)
  in

  Arg.current := 1;
  Arg.parse 
      [ "-build", Arg.String (word_by_word add_build_wish),
	       "<pkg>                 Build <pkg>";
	"-remove", Arg.String (word_by_word add_remove_wish),
	        "<pkg>                Remove <pkg>";
	"-add", Arg.String add_package,
	     "<tgzfile>               Add binary package <tgzfile>";
	"-option", Arg.String add_option,
	        "<pkg>:<name>=<val>   Set option <name>=<val> for <pkg>";
	"-wishes", Arg.Unit add_wishes,
	        "                     Build/remove packages of the wish list";
	"-rebuild", Arg.Set build_all_installed,
	         "                    Rebuild all installed packages";
	"-newer", Arg.Set newer,
	       "                      Install only newer packages";
	"-disable-builder", Arg.Set disable_builder,
	                 "            Do not try to build packages";
	"-auto-replace-old", Arg.Unit (fun _ -> replace := `Upgrade),
                          "           Try to upgrade old packages if necessary";
	"-dry-run", Arg.Set dry_run,
	         "                    Only generate a plan but don't carry it out";
      ]
      (fun _ -> raise (Arg.Bad "Unexpected argument"))
      "usage: godi_console perform <options>";

  (* First scan packages *)
  let src_pkgs = 
    if !disable_builder then [] else scan_source_packages() in
  let inst_pkgs = get_installed_packages() in
  let bin_pkgs = List.map get_binary_package !binary_additions in

  (* Apply wishes *)
  Godi_wishes.apply_wishes_to_packages src_pkgs !wishes;
  Godi_wishes.apply_wishes_to_packages inst_pkgs !wishes;
  List.iter (fun pkg -> (any_pkg pkg) # set_user_status `Add) bin_pkgs;

  let persistent_wl = Godi_wishes.read_wishes ~pref_user_error:true () in
  if !use_wishlist then (
    Godi_wishes.apply_wishes_to_packages src_pkgs persistent_wl;
    Godi_wishes.apply_wishes_to_packages inst_pkgs persistent_wl;
    (* wishlist does not support binary packages *)
  );
 
  if !build_all_installed then (
    let src_pkgs_map = PkgMap.make (List.map (fun pkg -> (pkg,())) src_pkgs) in
    List.iter
      (fun pkg ->
	 let n = pkg_name pkg in
	 let spkgs = PkgMap.find_by ~name:n src_pkgs_map in
	 List.iter
	   (fun (spkg,_) ->
	      (source_pkg spkg)#set_user_status `Build
	   )
	   spkgs
      )
      inst_pkgs
  );

  if !newer then (
    let inst_pkgs_map = PkgMap.make (List.map (fun pkg -> (pkg,())) inst_pkgs) in
    List.iter
      (fun pkg ->
	 let n = pkg_name pkg in
	 let ipkgs = PkgMap.find_by ~name:n inst_pkgs_map in
	 match ipkgs with
	     [] ->
	       ()
	   | [ipkg,_] ->
	       (* Is pkg newer than ipkg? *)
	       if not(newer_package pkg ipkg) then (
		 (* pkg is not newer, so do not build it *)
		 (any_pkg pkg)#set_user_status `Keep
	       )
	   | _ ->
	       (* Several installed packages with the same name: impossible *)
	       assert false
      )
      (src_pkgs @ bin_pkgs)
  );

  (* Prepare a function to write wishes back to the file: *)
  let save_wishes() =
    if !use_wishlist then (
      let wl = Godi_wishes.wishes_of_packages(src_pkgs @ inst_pkgs) in
      (* We only write those wishes that are also contained in 
       * persistent_wl. - Binary packages are out of scope.
       *)
      let wl' =
	List.filter (fun w -> List.mem w persistent_wl) wl in
      Godi_wishes.write_wishes ~pref_user_error:true wl';
    )
  in

  (* Write a temporary godi.conf file: *)
  let tmp_godi_conf =  Filename.temp_file "tmpgodi" ".conf" in
  let mflag = Godi_conffile.modify_godi_conf
                ~pref_user_error:true ~destfile:tmp_godi_conf src_pkgs in
 
  (* When options have been modified, rescan source packages: *)
  let vid = new_var_id() in
  let src_pkgs' =
    if mflag then (
      List.iter (fun pkg -> 
		   (source_pkg pkg)#rescan_deps tmp_godi_conf vid)
		src_pkgs;
      List.map 
	(fun pkg -> `Source_pkg ((source_pkg pkg)#variant vid))
	src_pkgs
    )
    else src_pkgs in
  Sys.remove tmp_godi_conf;

  (* Generate the plan: *)
  let _n = Godi_plan.update_plan 
	    ~replace:!replace 
	    ~iteration:1 
	    (src_pkgs' @ inst_pkgs @ bin_pkgs) in
  (* May raise Mission_impossible, or Package_installation_not_unique *)
 
  let to_remove =
    List.filter to_be_removed inst_pkgs in

  let to_remove' =
    List.rev(Godi_plan.sort_packages to_remove) in
 
  let to_build_or_add =
    List.filter to_be_built src_pkgs' @
    List.filter to_be_added bin_pkgs in

  let to_build_or_add' =
    Godi_plan.sort_packages to_build_or_add in

  (* Execute the plan: *)

  if to_remove = [] && to_build_or_add = [] && not mflag then
    print_endline "Nothing to do."
  else (
    List.iter
	  (fun pkg ->
	     if pkgtype pkg = `Source then (
	       printf "### Fetching %s (if necessary)\n" (pkg_name pkg);
	       flush stdout;
	       if not !dry_run then
		 Godi_perform.fetch_pkg ~pref_user_error:true pkg;
	     )
	   )
	   to_build_or_add';
    List.iter
	  (fun pkg ->
	     if (installed_pkg pkg)#preserve then
	       printf "### Package %s cannot be removed, skipping\n" (pkg_fullname pkg)
	     else (
	       printf "### Removing %s\n" (pkg_fullname pkg);
	       flush stdout;
	       if not !dry_run then (
		 ( try
		     Godi_perform.remove_pkg ~soft_errors:true pkg
		   with
		       Godi_perform.Soft_rm_error ->
			 ()
		 );
		 (installed_pkg pkg)#set_user_status `Keep;
		 save_wishes();
	       )
	     )
	  )
	  to_remove';

    if mflag then (
      printf "### Modifying godi.conf\n";
      flush stdout;
      if not !dry_run then (
	let _ = 
	  Godi_conffile.modify_godi_conf ~pref_user_error:true to_build_or_add' 
	in ()
      )
    );

    List.iter
      (fun pkg ->
	 match pkgtype pkg with
	   | `Source ->
	       printf "### Building %s\n" (pkg_name pkg);
	       flush stdout;
	       if not !dry_run then (
		 try
		   Godi_perform.build_pkg ~pref_user_error:true pkg;
		   (any_pkg pkg)#set_user_status `Keep;
		   save_wishes();
		 with
		     Restart_recommended sh_cmd ->
		       printf "### The installation of the package requires a restart of this program!\n";
		       flush stdout;
		       (* Remove the wish, if any. Note that we do not know whether
			* sh_cmd will be successful, so this may be too early.
			*)
		       (any_pkg pkg)#set_user_status `Keep;
		       save_wishes();
		       (* Determine the remaining packages to build: *)
		       let to_do = ref [] in
		       let found = ref false in
		       List.iter (fun pkg' -> 
				    if !found then to_do := pkg' :: !to_do;
				    if pkg = pkg' then found := true) 
			         to_build_or_add';
		       let restart_args =
			 List.map
			   (fun pkg' ->
			      match pkg' with
				  `Source_pkg _ ->
				    "-build " ^ Filename.quote (pkg_name pkg')
				| `Binary_pkg p ->
				    "-add " ^ Filename.quote p#binary_archive
				| _ -> assert false
			   )
			   (List.rev !to_do) in
		       let restart_cmd =
			 "exec " ^ Filename.quote Sys.argv.(0) ^ " perform " ^ 
			 String.concat " " restart_args in
		       raise(Restart(sh_cmd ^ "\n" ^ restart_cmd));
		   
	       )
	   | `Binary ->
	       printf "### Adding %s\n" (pkg_fullname pkg);
	       flush stdout;
	       if not !dry_run then (
		 Godi_perform.add_pkg pkg;
		 (any_pkg pkg)#set_user_status `Keep;
		 save_wishes();
	       )
	   | `Installed | `Dispersed ->
	       assert false
      )
      to_build_or_add'
  )
;;


let print_deps l =
  String.concat ", "
    (List.map Godi_versions.print_dependency l)
;;


let indent s =
  Str.global_replace (Str.regexp "\n") "\n    " s
;;


let remove_linefeeds s =
  Str.global_replace (Str.regexp "[\r\n]") " " s
;;


let version_str pkg =
  let v = pkg_version pkg in
  let r = pkg_pkgrevision pkg in
  if r > 0 then v ^ "godi" ^ string_of_int r else v
;;


let cmd_list() =
  let list_src = ref false in
  let list_inst = ref false in
  let only_names = ref false in
  let long = ref false in
  let filters = ref [] in

  Arg.current := 1;
  Arg.parse 
      [ "-source", Arg.Set list_src,
                "           List packages available as source code";
	"-installed", Arg.Set list_inst,
	           "        List installed packages";
	"-names", Arg.Set only_names,
	       "            Output the pure list of package names (no versions etc.)";
	"-long", Arg.Set long,
	      "             Output descriptions, options, etc.";
	
      ]
      (fun s -> 
	 let dep = Godi_versions.parse_dependency s in
	 filters := !filters @ [dep])
      "usage: godi_console list [<options>] [<pkgfilter> ...]";

  if not !list_src && not !list_inst then (
    list_src := true;
    list_inst := true;
  );

  (* First scan packages *)
  let src_pkgs = scan_source_packages() in
  let inst_pkgs = get_installed_packages() in
  let out_pkgs = src_pkgs @ inst_pkgs in

  (* filters: Select which packages are printed *)
  let is_selected pkg =
    (!filters = []) ||
    ( (* Package is selected when there is a matching filter *)
      List.exists
	(fun (n,cond) ->
	   (n = (pkg_name pkg)) &&
	   (Godi_versions.version_matches_condition
	      (pkg_version pkg) cond))
	!filters
    )
  in

  (* Arrange packages by name: *)
  let h = Hashtbl.create 20 in
  let keys = ref [] in
  List.iter
    (fun pkg ->
       let n = pkg_name pkg in
       let old =
	 try Hashtbl.find h n with Not_found -> [] in
       if old = [] then keys := n :: !keys;
       Hashtbl.replace h n (pkg :: old)
    )
    out_pkgs;

  keys := List.sort compare !keys;

  (* Output list *)
  List.iter
    (fun name ->
       let pkgs = Hashtbl.find h name in
       let (inst_version, inst_sel) =
	 try
	   let inst_pkg = 
	     List.find (fun pkg -> pkgtype pkg = `Installed) pkgs in
	   ( Some(version_str inst_pkg),
	     !list_inst && is_selected inst_pkg
	   )
	 with
	     Not_found -> (None, false) in
       let (src_version, src_sel) =
	 try
	   let src_pkg = 
	     List.find (fun pkg -> pkgtype pkg = `Source) pkgs in
	   ( Some(version_str src_pkg),
	     !list_src && is_selected src_pkg
	   )
	 with
	     Not_found -> (None, false) in

       let sel = inst_sel || src_sel in
       
       if sel then (
	 if !only_names then
	   print_endline name
	 else (
	   let inst_version_str =
	     match inst_version with
		 None -> "-"
	       | Some v -> v in
	   let src_version_str =
	     match src_version with
		 None -> "-"
	       | Some v -> v in

	   (* Long format:
	    * - The package name is on column 0
	    * - The attributes begin on column 2
	    * - Continuation lines begin on column 4
	    * ==> Easy to parse!
	    *)

	   if !long then (
	     List.iter
               (fun pkg ->			    
		  let sel =
		    match pkgtype pkg with
			`Source -> !list_src
		      | `Installed -> !list_inst
		      | _ -> false in
		  if sel then (
		    printf "%s\n" name;
		    ( match pkgtype pkg with
			  `Source -> 
			    printf "  type: source\n"
			| `Installed ->
			    printf "  type: installed\n"
			| _ -> assert false
		    );
		    printf "  version: %s\n" (version_str pkg);
		    printf "  comment: %s\n" (remove_linefeeds (pkg_comment pkg));
		    printf "  description: %s\n" (indent (pkg_description pkg));
		    printf "  dependencies: %s\n" (print_deps
						     (pkg_dependencies pkg));
		    printf "  build_deps: %s\n" (print_deps
						   (pkg_build_dependencies pkg));
		    ( match pkgtype pkg with
			  `Source ->
			    printf "  build_message: %s\n"
			      (indent (source_pkg pkg)#build_message);
			    let nv = (source_pkg pkg)#values_of_build_options in
			    printf "  build_options: %s\n"
			      (indent
				 (String.concat "\n" 
				    (List.map
				       (fun (n,v) -> n ^ "=" ^ v)
				       nv)));
			| `Installed ->
			    printf "  preserved: %b\n"
			      (installed_pkg pkg)#preserve;
			| _ ->
			    ()
		    );
		    printf "\n";
		  )
	       )
	       pkgs

	   )
	   else (
	     printf "%-30s inst:%-10s src:%-10s\n"
		   name
		   inst_version_str
		   src_version_str
	   )
	 )
       )
    )
    !keys;
;;


let cmd_dep() =
  let mode = ref `Normal in
  let query_type = ref `Names in
  let dep_expressions = ref [] in

  Arg.current := 1;
  Arg.parse
    [ "-quiet", Arg.Unit (fun () -> mode := `Quiet),
      "  Only output whether the dependencies are fulfilled but not how";
      
      "-verbose", Arg.Unit (fun () -> mode := `Verbose),
      "  Output the result in a verbose style";
      
      "-names", Arg.Unit (fun () -> query_type := `Names),
      "  Print the actually matching pkg names with version numbers for every dep";
      
      "-names-all", Arg.Unit (fun () -> query_type := `Names_all),
      "  Print the matching pkg names+versions for all deps including prerequisites";
      
      "-size", Arg.Unit (fun () -> query_type := `Size),
      "  Print the package size in bytes for every dep";
      
      "-size-all", Arg.Unit (fun () -> query_type := `Size_all),
      "  Print the summed size of the matching pkgs and all prerequisites in bytes";
    ]
    (fun s -> dep_expressions := !dep_expressions @ [s])
    "usage: godi_console dep [<options>] <pgk-dep> ...";

  let inst_pkgs = Godi_query.get_installed_packages () in
  (* Read packages now! *)

  let success = ref true in

  let pkg_opt_list =
    (* The list of installed packages corresponding to dep_expressions *)
    List.map
      (fun dep ->
	 let (name, cond) = 
	   try Godi_versions.parse_dependency dep
	   with Failure msg -> raise (User_error msg)
	 in

	 try
	   let pkg = Godi_query.get_this_installed_package name in 
	   (* or Not_found *)
	   if Godi_query.matches_dep pkg (name, cond) then
	     (dep, Some pkg)
	   else
	     (dep, None)
	 with
	     Not_found -> (dep, None)
      )
      !dep_expressions in

  (match !query_type with
     | `Names
     | `Size ->
	 (* Process every dependency separately! *)
	 List.iter
	   (fun (dep, pkg_opt) ->
	      success := !success && (pkg_opt <> None);
	      match !mode with
		| `Quiet -> ()
		| `Normal ->
		    ( match pkg_opt with
			| None -> 
			    eprintf "%s: not found\n" dep; flush stderr
			| Some pkg ->
			    ( match !query_type with
				| `Names ->
				    printf "%s\n" (pkg_fullname pkg); 
				    flush stdout
				| `Size ->
				    let ipkg = installed_pkg pkg in
				    printf "%Ld\n" (ipkg # size); flush stdout
				| _ -> assert false
			    )
		    )
		| `Verbose ->
		    ( match pkg_opt with
			| None -> 
			    eprintf "%s: not found\n" dep; flush stderr
			| Some pkg ->
			    let n = pkg_fullname pkg in
			    ( match !query_type with
				| `Names ->
				    printf "%s: %s\n" dep n; flush stdout
				| `Size ->
				    let ipkg = installed_pkg pkg in
				    printf "%s: %Ld\n" n (ipkg # size); 
				    flush stdout
				| _ -> assert false
			    )
		    )
	   )
	   pkg_opt_list

     | `Names_all
     | `Size_all ->
	 (* First compute the closure of the set of packages: *)
	 let pkgs =
	   List.flatten
	     (List.map
		(fun (dep, pkg_opt) ->
		   success := !success && (pkg_opt <> None);
		   match pkg_opt with
		     | None ->
			 eprintf "%s: not found\n" dep; flush stderr;
			 []
		     | Some pkg ->
			 [pkg]
		)
		pkg_opt_list 
	     ) in

	 let reqs =
	   Godi_plan.get_required_packages
	     ~active:inst_pkgs
	     ~selected:pkgs
	     () in

	 let all = 
	   Godi_plan.sort_packages (pkgs @ reqs) in

	 (* all: The packages mentioned in the dep expressions plus all
          * prerequisites, sorted topologically.
          *)

	 let total_size() =
	   List.fold_left
	     (fun acc pkg -> Int64.add acc ((installed_pkg pkg) # size))
	     0L
	     all in

	 match !mode with
	   | `Quiet -> ()
	   | `Normal ->
	       ( match !query_type with
		   | `Names_all ->
		       List.iter
			 (fun pkg ->
			    printf "%s\n" (pkg_fullname pkg); 
			    flush stdout
			 )
			 all
		   | `Size_all ->
		       printf "%Ld\n" (total_size());
		       flush stdout
		   | _ ->
		       assert false
	       )
	       
	   | `Verbose ->
	       ( match !query_type with
		   | `Names_all ->
		       List.iter
			 (fun pkg ->
			    printf "%s\n" (pkg_fullname pkg); 
			    flush stdout
			      (* Don't know how to be verbose here! *)
			 )
			 all
		   | `Size_all ->
		       printf "Total size: %Ld\n" (total_size());
		       flush stdout
		   | _ ->
		       assert false
	       )
  );

  if not !success then
    raise (Exit_code 1)
;;


let cmd_delete() =
  let mode = ref `Normal in
  let recurse = ref false in
  let verbose = ref false in
  let package_names = ref [] in

  Arg.current := 1;
  Arg.parse
      [ "-f", Arg.Unit (fun () -> 
			  if !mode <> `Normal then 
			    raise(Arg.Bad "Incompatible switches");
			  mode := `Force),
	   "   Force deletion";
	"-n", Arg.Unit (fun () -> 
			  if !mode <> `Normal then 
			    raise(Arg.Bad "Incompatible switches");
			  mode := `Simulate),
	   "   Only simulate deletion";
	"-r", Arg.Set recurse,
	   "   Delete successor packages";
	"-v", Arg.Set verbose,
	   "   Turn on verbose output";
      ]
      (fun s -> package_names := !package_names @ [s])
      "usage: godi_console delete [<options>] <pkg>...";

  (* -f and -r are incompatible: Too risky *)
  if !mode = `Force && !recurse then
    user_error("The options -f and -r are incompatible with each other");

  (* General algorithm:
   *
   * The listed packages are looked up, and the package objects are passed
   * to the planner. The result is that the planner suggests to delete all
   * successor packages, and the listed packages. If -r, we can just do that.
   * If not -r, and there are packages not listed on the command line, stop
   * with an error message.
   *
   * The option -f has two effects: First, it is now allowed that the planner
   * returns successor packages. These are not deleted, however, but just
   * ignored. The [mode] argument is passed to [remove_pkg], so the following
   * is turned off:
   * - Checksum test
   * - Requirements script
   *)
  let protocol = `Forward (fun s -> prerr_string s; flush stderr) in
  let message = Godi_exec.collect protocol in
  let inst_pkgs = get_installed_packages ~protocol ~verbose:!verbose() in

  let sel_pkgs =
    List.map
      (fun s ->
	 (* If the package name has a version string, we accept that. It is
	  * then tested that the version matches.
	  *)
	 inst_pkg_by_name s
      )
      !package_names in
  List.iter
    (fun pkg -> (installed_pkg pkg) # set_user_status `Remove)
    sel_pkgs;
  let _sel_pkgs_set = PkgMap.make (List.map (fun p -> (p,())) sel_pkgs) in

  if !verbose then
    message "Planning the removal\n";
  let n = Godi_plan.update_plan 
	    ~iteration:1 
	    inst_pkgs in
  (* May raise Mission_impossible, or Package_installation_not_unique *)

  if n>0 && not !recurse && !mode <> `Force then
    user_error("There are dependent packages, cannot delete");

  let to_remove =
    List.filter to_be_removed inst_pkgs in

  let to_remove' =
    List.rev(Godi_plan.sort_packages to_remove) in

  let had_errors = ref false in

  if to_remove = [] then
    ( if !verbose then 
	message "Nothing to do.\n" )
  else (
    List.iter
	  (fun pkg ->
	     if (installed_pkg pkg)#preserve && !mode <> `Force then
	       message
	         (sprintf "Package %s cannot be removed, skipping\n" (pkg_fullname pkg))
	     else (
	       let do_remove =
		 match pkg_user_status pkg with
		     `Remove -> true
		   | `Remove_dep _ -> !recurse
		   | _ -> false 
	       in
	       if do_remove then (
		 if !verbose then
		   message
		     (sprintf "Removing %s\n" (pkg_fullname pkg));
		 ( try
		     Godi_perform.remove_pkg 
		       ~soft_errors:true ~protocol ~mode:!mode pkg;
		   with
		       Godi_perform.Soft_rm_error ->
			 had_errors := true
		 )
	       )
	     )
	  )
	  to_remove';
  );
  if !had_errors then
    raise(Exit_code 1)
;;


let cmd_add() =
  let mode = ref `Normal in
  let upgrade = ref false in
  let verbose = ref false in
  let package_files = ref [] in

  Arg.current := 1;
  Arg.parse
      [ "-u", Arg.Set upgrade,
	   "   Replace installed package";
	"-n", Arg.Unit (fun () -> mode := `Simulate),
	   "   Only simulate addition";
	"-v", Arg.Set verbose,
	   "   Turn on verbose output";
	(* Later:
	 * -pool <location>: Take package files from this pool. Also
	 *    GODI_BINARY_POOLS environment variable for the same purpose.
	 * -r: Add missing requirements automatically
	 *
	 * Idea:
	 * -f: We can force the deletion (in conjuction with -u)
	 *)
      ]
      (fun s -> package_files := !package_files @ [s])
      "usage: godi_console add [<options>] <pkgfile.tgz>...";

  let protocol = `Forward (fun s -> prerr_string s; flush stderr) in
  let message = Godi_exec.collect protocol in
  let inst_pkgs = get_installed_packages ~protocol ~verbose:!verbose() in

  if !verbose then
    message "Scanning files to install...\n";
  
  let binary_pkgs =
    List.map get_binary_package !package_files in
  
  List.iter
    (fun pkg -> (binary_pkg pkg) # set_user_status `Add)
    binary_pkgs;

  if !verbose then
    message "Planning the addition\n";
  let n = Godi_plan.update_plan 
	    ~iteration:1 
	    (inst_pkgs @ binary_pkgs) in
  (* May raise Mission_impossible, or Package_installation_not_unique *)

  if n>0 && not !upgrade then
    user_error("Some packages are already installed, cannot add");

  let to_remove =
    List.filter to_be_removed inst_pkgs in

  let to_remove' =
    List.rev(Godi_plan.sort_packages to_remove) in
 
  let to_add =
    List.filter to_be_added binary_pkgs in

  let to_add' =
    Godi_plan.sort_packages to_add in

  (* Execute the plan: *)

  let had_errors = ref false in

  if to_remove = [] && to_add = [] then
    ( if !verbose then 
	message "Nothing to do.\n" )
  else (
    List.iter
	  (fun pkg ->
	     if (installed_pkg pkg)#preserve then
	       message
	         (sprintf "Package %s cannot be removed, skipping\n" (pkg_fullname pkg))
	     else (
	       if !verbose then
		 message
		   (sprintf "Removing %s\n" (pkg_fullname pkg));
	       ( try
		   Godi_perform.remove_pkg 
		     ~soft_errors:true ~protocol 
		     ~mode:(!mode :> Godi_perform.rm_mode) pkg;
		 with
		     Godi_perform.Soft_rm_error ->
		       had_errors := true
	       )
	     )
	  )
	  to_remove';

    List.iter
      (fun pkg ->
	 if !verbose then
	   message (sprintf "Adding %s\n" (pkg_fullname pkg));
	 Godi_perform.add_pkg ~mode:!mode ~protocol pkg;
      )
      to_add'
  );
  if !had_errors then
    raise(Exit_code 1)
;;



let colon_re = Str.regexp ":.*$";;


let cmd_ptool() =
  let input = ref None in
  let output = ref None in

  let prefix = ref None in
  let dispersion = ref None in

  let fixup_symlinks = ref false in

  let disp = ref false in
  let disp_name = ref None in
  let disp_comment = ref None in
  let disp_description = ref None in
  let disp_maintainer = ref "" in
  let disp_dependencies = ref [] in
  let disp_build_dependencies = ref [] in
  let disp_conflicts = ref [] in
  let disp_install_message = ref "" in
  let disp_build_version = ref "" in
  let disp_build_info = ref "" in
  let disp_preserve = ref false in
  let disp_req_script = ref None in
  let disp_install_script = ref None in
  let disp_deinstall_script = ref None in

  let parse_dep_list s =
    let l = split s in
    List.map (fun  s -> 
		(* Strip ":"-part off *)
		let s' = Str.global_replace colon_re "" s in
		Godi_versions.parse_dependency s') l
  in

  let only_one_input() =
    if !input <> None then
      user_error "It is only possible to have one input"
  in

  let only_one_output() =
    if !output <> None then
      user_error "It is only possible to have one output"
  in


  Arg.current := 1;
  Arg.parse
      [ "-in-installed",
	Arg.String (fun s -> only_one_input(); input := Some(`Installed s)),
	"<pkg-name>  Process this installed package";

	"-in-dispersed",
	Arg.Unit (fun () -> only_one_input(); input := Some `Dispersed),
	"  Process dispersed files characterized by -set-* options";

	"-fixup-symlinks",
	Arg.Set fixup_symlinks,
	"  Modify the input, and fix up non-relative symlinks";

	"-out-register",
	Arg.Unit (fun () -> only_one_output(); output := Some `Register),
	"  Register the input in the GODI package database";

	"-out-install",
	Arg.Unit (fun () -> only_one_output(); output := Some `Install),
	"  Install the input as GODI package (scripts are not run)";

	"-out-binary",
	Arg.String (fun s -> only_one_output(); output := Some (`Binary s)),
	"<dir>  Create a binary package in the directory <dir>";

	"-set-prefix",
	Arg.String (fun s -> disp := true; prefix := Some s),
	"<prefix>  Take dispersed files from this directory hierarchy";

	"-set-plist",
	Arg.String (fun s -> disp := true; dispersion := Some(`Plist s)),
	"<file>  Select dispersed files matching this PLIST <file>";

	"-set-auto-plist",
	Arg.Unit (fun () -> disp := true; dispersion := Some `Auto_plist),
	"  Select all files in <prefix> (autogenerated PLIST)";
	
	"-set-name",
	Arg.String (fun s -> disp := true; disp_name := Some s),
	"<pkg-name>  Set the name of the package";

	"-set-comment",
	Arg.String (fun s -> disp := true; disp_comment := Some s),
	"<text>  Set the comment string of the package";

	"-set-description",
	Arg.String (fun s -> disp := true; disp_description := Some s),
	"<text>  Set the description text of the package";

	"-set-description-file",
	Arg.String (fun s -> 
		      disp := true; 
		      disp_description := 
			Some(read_text_file ~pref_user_error:true s)),
	"<file>  Take the description text of the package from <file>";

	"-set-maintainer",
	Arg.String (fun s -> disp := true; disp_maintainer := s),
	"<text>  Set the name of the GODI maintainer";

	"-set-dependencies",
	Arg.String (fun s -> 
		      disp := true; disp_dependencies := parse_dep_list s),
	"<deps>  Set the (space-seperated) dependencies";

	"-set-build-dependencies",
	Arg.String (fun s -> 
		      disp := true; 
		      disp_build_dependencies := parse_dep_list s),
	"<deps>  Set the (space-seperated) build dependencies";

	"-set-conflicts",
	Arg.String (fun s -> disp := true; disp_conflicts := parse_dep_list s),
	"<deps>  Set the (space-seperated) conflicts";

	"-set-install-message",
	Arg.String (fun s -> disp := true; disp_install_message := s),
	"<text>  Set the message printed on installation";

	"-set-install-message-file",
	Arg.String (fun s -> 
		      disp := true; 
		      disp_install_message := 
			(read_text_file ~pref_user_error:true s)),
	"<file>  Take the message printed on installation from <file>";

	"-set-build-version",
	Arg.String (fun s -> disp := true; disp_build_version := s),
	"<text>  Set the build version text";

	"-set-build-version-file",
	Arg.String (fun s -> 
		      disp := true; 
		      disp_build_version := 
			(read_text_file ~pref_user_error:true s)),
	"<file>  Take the build version text from <file>";

	"-set-build-info",
	Arg.String (fun s -> disp := true; disp_build_info := s),
	"<text>  Set the build info text";

	"-set-build-info-file",
	Arg.String (fun s -> 
		      disp := true; 
		      disp_build_info := 
			(read_text_file ~pref_user_error:true s)),
	"<file>  Take the build info text from <file>";

	"-set-preserve",
	Arg.Bool (fun b -> disp := true; disp_preserve := b),
	"<bool>  Set whether the package is protected from being deleted";

	"-set-req-script",
	Arg.String (fun s -> disp := true; disp_req_script := Some s),
	"<file>  Set the requirements script";

	"-set-install-script",
	Arg.String (fun s -> disp := true; disp_install_script := Some s),
	"<file>  Set the install script";

	"-set-deinstall-script",
	Arg.String (fun s -> disp := true; disp_deinstall_script := Some s),
	"<file>  Set the deinstall script";
      ]
    (fun s -> raise(Arg.Bad "Unexpected argument"))
    "Usage: godi_console ptool <option>...";

  if !input = None then
    user_error "Missing input";

  if !output = None then
    user_error "Missing output";

  if !input <> Some `Dispersed && !disp then
    user_error "The -set-* options are only allowed for -in-dispersed";

  let in_pkg =
    match !input with
      | None -> user_error "Missing input"
      | Some (`Installed pkg_name) ->
	  let pkg = inst_pkg_by_name pkg_name in
	  Godi_ptool.dispersed_of_installed_package pkg
      | Some `Dispersed ->
	  let need_arg var name =
	    match !var with
	      | None -> user_error ("Missing parameter: " ^ name)
	      | Some v -> v
	  in
	  let full_name = need_arg disp_name "-set-name" in
	  let (name, version, pkgrevision) = 
	    Godi_versions.split_name_version full_name in
	  
	  let prefix = need_arg prefix "-set-prefix" in

	  let d =
	    match !dispersion with
	      | None ->
		  user_error "Need either -set-plist or -set-auto-plist"
	      | Some (`Plist f) ->
		  let plist = Godi_plist.parse_user_plist f in
		  Godi_ptool.user_plist_dispersion ~root:prefix plist
	      | Some `Auto_plist ->
		  Godi_ptool.stage_dispersion prefix
	  in

	  Godi_ptool.create_dispersed_package
	    ~name ~version ~pkgrevision
	    ~comment:(need_arg disp_comment "-set-comment")
	    ~description:(need_arg disp_description "-set-description")
	    ~godi_maintainer:!disp_maintainer
	    ~dependencies:!disp_dependencies
	    ~build_dependencies:!disp_build_dependencies
	    ~conflicts:!disp_conflicts
	    ~install_message:!disp_install_message
	    ~build_version:!disp_build_version
	    ~build_info:!disp_build_info
	    ~preserve:!disp_preserve
	    ~requirements_script:!disp_req_script
	    ~install_script:!disp_install_script
	    ~deinstall_script:!disp_deinstall_script
	    d
  in

  if !fixup_symlinks then
    Godi_ptool.fixup_symlinks (dispersed_pkg in_pkg)#dispersion;

  match !output with
    | None -> assert false
    | Some `Register ->
	Godi_perform.register_pkg in_pkg
    | Some `Install ->
	Godi_perform.install_pkg in_pkg
    | Some (`Binary dir) ->
	let name = pkg_name in_pkg in
	let version = pkg_version in_pkg in
	let pkgrevision = pkg_pkgrevision in_pkg in
	let out_file_name =
	  (Filename.concat 
	     dir
	     (Godi_versions.concat_name_version name version pkgrevision))
	    ^ ".tgz" in
	let out_pkg =
	  Godi_ptool.binary_of_dispersed_package out_file_name in_pkg in
	printf "Wrote %s\n" (binary_pkg out_pkg)#binary_archive
;;
	




let assign_re = Str.regexp "^\\([^=]+\\)=\\(.*\\)$";;


let cmd_make argi =
  (* "make" is special, because we need to look at $MAKEFLAGS for additional
   * arguments 
   *)

  let args_makeflags =
    try
      let makeflags = Sys.getenv "MAKEFLAGS" in
      Array.of_list(Godi_match.parse_shell_words makeflags)
    with
	Not_found -> [| |] in
  let args_godi_makeflags =
    try
      let makeflags = Sys.getenv "GODI_MAKEFLAGS" in
      Array.of_list(Godi_match.parse_shell_words makeflags)
    with
	Not_found -> [| |] in
  let args_argv =
    (* Without command and "make" *)
    Array.sub Sys.argv argi (Array.length Sys.argv - argi) in
  let args =
    Array.concat [ [| "godi_make" |];    (* do this always! *)
		   args_godi_makeflags; 
		   args_makeflags; 
		   args_argv ] in
  (* Because later arguments oveeride earlier ones, it is quite important
   * that [args_argv] is evaluated after [args_makeflags], so directly
   * passed arguments have precedence.
   *)

  let level =
    try
      int_of_string (Sys.getenv "MAKELEVEL")
    with
      | Failure _
      | Not_found -> 1
  in

  let makefile = ref None in

  if Sys.file_exists "makefile" then
    makefile := Some "makefile"
  else
    if Sys.file_exists "Makefile" then
      makefile := Some "Makefile";

  let inc_dirs = ref [ "." ] in
  let sys_inc_dirs = ref [] in
  let global_ignore_flag = ref false in
  let global_silent_flag = ref false in
  let exec_mode = ref `Execute in
  let mode = ref `Default in
  let vars = ref [] in
  let override = ref [] in
  let targets = ref [] in
  let sys_mk = ref true in
  let options = ref [] in
  let chdir = ref None in
  let trace_flag = ref false in
  let trace_script_flag = ref false in

  Arg.current := 0;
  Arg.parse_argv
    args
    [ "-f", Arg.String (fun s -> makefile := Some s),
         "<file>  Set the Makefile to read";
      "-C", Arg.String (fun s -> chdir := Some s),
         "<dir>   Change into directory";
      "-I", Arg.String (fun s -> inc_dirs := !inc_dirs @ [s];
			         options := !options @ ["-I"; s]
		       ),
         "<dir>   Search in this directory";
      "-m", Arg.String (fun s -> sys_inc_dirs := !sys_inc_dirs @ [s];
			         options := !options @ ["-m"; s]
		       ),
         "<dir>   Search in this system directory";
      "-i", Arg.Unit (fun () -> global_ignore_flag := true;
			        options := !options @ ["-i"]
		     ),
         "        Ignore failed commands";
      "-n", Arg.Unit (fun () -> exec_mode := `Simulate_except;
			        options := !options @ ["-n"]
		     ),
         "        Simulate execution (except .MAKE targets)";
      "-N", Arg.Unit (fun () -> exec_mode := `Simulate;
			        options := !options @ ["-N"]
		     ),
         "        Simulate execution";
      "-q", Arg.Unit (fun () -> exec_mode := `Test;
			        options := !options @ ["-q"]
		     ),
         "        Only test whether targets are up to date";
      "-r", Arg.Unit (fun () -> sys_mk := false;
			        options := !options @ ["-r"]
		     ),
	 "        Do not include <sys.mk>";
      "-s", Arg.Unit (fun () -> global_silent_flag := true;
			        options := !options @ ["-s"]
		     ),
         "        Be silent";
      "-t", Arg.Unit (fun () -> exec_mode := `Fake;
			        options := !options @ ["-t"]
		     ),
         "        Fake the update of targets";
      "-trace", Arg.Unit (fun () -> trace_flag := true;
			            options := !options @ [ "-trace" ]),
             "    Trace the execution";
      "-trace-script", Arg.Unit (fun () -> trace_script_flag := true;
			            options := !options @ [ "-trace-script" ]),
             "    Call the shell with -x";
      "-V", Arg.String (fun s -> 
			  if !mode <> `Default && !mode <> `Variable then
			    raise(Arg.Bad "-V incompatible with specifying a target");
			  mode := `Variable;
			  vars := !vars @ [s]),
         "<var>   Output the unexpanded definition of <var>";
    ]
    (fun arg ->
       if Str.string_match assign_re arg 0 then
	 override := !override @ [ Str.matched_group 1 arg,
				   Str.matched_group 2 arg ]
       else (
	 if !mode <> `Default && !mode <> `Target then
	   raise(Arg.Bad "-V incompatible with specifying a target");
	 mode := `Target;
	 targets := !targets @ [ arg ];
       )

    )
    "usage: godi_console make <options> ( var=value | target ) ...";

  (match !chdir with
     | None -> ()
     | Some d -> Unix.chdir d
  );

  let ctx =
    Godi_make.create_context
      ~override:!override
      ~options:(Godi_match.quote_shell_words !options)
      ~user_mk_path:!inc_dirs
      ?sys_mk_path:(if !sys_inc_dirs = [] then None else Some !sys_inc_dirs)
      ~make_cmd:"godi_make"
      ~trace:!trace_flag
      ~trace_script:!trace_script_flag
      !targets
  in

  let pcache = Godi_make.create_parser_cache() in

  let make_user_file =
    if !exec_mode = `Execute then
      Some (fun file ctx ->
	      let ctx' = { ctx with Godi_make.final_targets = [file] } in
	      Godi_make.execute
		~exec_mode:!exec_mode
		level
		ctx')
    else
      None
  in

  Godi_make.fast_parse_eval
    ~pref_user_error:true
    ~sys_mk:!sys_mk
    ?make_user_file
    pcache
    ( match !makefile with
	| None -> user_error "No Makefile available"
	| Some f -> f
    )
    ctx;

  if !global_ignore_flag then
    ctx.Godi_make.global_ignore_flag <- true;
  if !global_silent_flag then
    ctx.Godi_make.global_silent_flag <- true;

  ( match !mode with
      | `Variable ->
	  List.iter
	    (fun varname ->
	       print_endline (Godi_make.var_definition ctx varname)
	    )
	    !vars
      | `Default | `Target ->
	  Godi_make.execute
	    ~pref_user_error:true
	    ~exec_mode:!exec_mode
	    level
	    ctx
  )
;;


let exec_cmd() =
  try
    ( let n = Array.length Sys.argv in
      if n < 2 then usage();
      match Sys.argv.(1) with
	  "update"  -> cmd_update()
	| "wish"    -> cmd_wish()
	| "perform" -> cmd_perform()
	| "list"    -> cmd_list()
	| "dep"     -> cmd_dep()
	| "delete"  -> cmd_delete()
	| "add"     -> cmd_add()
	| "ptool"   -> cmd_ptool()
	| "make"    -> cmd_make 2
	| _ ->
	    usage()
    );
  with
    | Godi_plan.Mission_impossible text ->
	prerr_endline "There is no way to perform the task.";
	prerr_endline text;
	raise (Exit_code 1)
    | Godi_plan.Package_installation_not_unique name ->
	prerr_endline ("Tried to install several versions of this package at once: " ^ name);
	raise (Exit_code 1)
    | User_error msg ->
	prerr_endline ("Error: " ^ msg);
	raise (Exit_code 1)
    | Godi_plist.Plist_error (file, line) ->
	prerr_endline ("Error in PLIST file " ^ file ^ ", in line:\n" ^ line)
    | Godi_plist.Plist_processing_error (file, text) ->
	prerr_endline ("Error in PLIST file " ^ file ^ ": " ^ text)
    | Godi_plist.Plist_expansion_error text ->
	prerr_endline ("Error during PLIST expansion: " ^ text)
    | Godi_make.Out_of_date ->
	(* Silent! *)
	raise (Exit_code 1)
    | Arg.Bad msg ->
	prerr_endline msg;
	raise (Exit_code 1)
    | Arg.Help msg ->
	prerr_endline msg;
	raise (Exit_code 0)
;;
