(* $Id: godi_plan.mli 678 2007-02-25 17:16:54Z gerd $
 * ----------------------------------------------------------------------
 * Copyright 2003 by Gerd Stolpmann. 
 * 
 * This file is part of godi-tools. 
 *
 * godi-tools is free software; you can redistribute it and/or modify it
 * under the conditions in the file LICENSE.
 *
 * godi-tools is distributed WITHOUT ANY WARRANTY, as detailed in the
 * file LICENSE.
 *)

(* Analyze dependencies and plan the build process *)

open Godi_types
open Godi_exec

exception Mission_impossible of string
  (** The string arguments describes the most likely reason. *)

exception Package_installation_not_unique of string
  (** There two source or binary packages with the
   * same name that are both selected for build, or for addition,
   * respectively. The package name is the argument of the exception.
   *)

type replacement_strategy =
    [ `Never
    | `Upgrade
    | `Any
    ]
  (** How to handle the case that there is an alternate package for an 
   * already installed package:
   * - [`Never]: The installed package is always kept as it is
   * - [`Upgrade]: Installed packages may be replaced by newer packages
   * - [`Any]: Installed packages may be replaced by packages of any
   *   version
   *)

val analyze_requirements : 
      ?replace:replacement_strategy ->
      available:package list ->       (* Which packages are available *)
      selected:package list ->        (* Which packages are already selected *)
      unit ->
	package list
      (** Checks which packages must be also selected to match the requirements
       * of the selected packages.
       * The returned package list is the list of missing requirements that
       * have been found in the list of available packages (including the
       * already selected packages). The returned list always contains
       * all members of [selected] (but see the option [replace] below), plus
       * the required members of
       * [available]. If there is no solution for this problem, the
       * exception [Mission_impossible] will be raised.
       *
       * All packages can be source packages, binary packages or installed
       * packages.
       *
       * The package names in [selected] must be unique (several versions of
       * the same package are not allowed).
       *
       * The option [replace] selects whether it is allowed that selected
       * packages are replaced by other versions found in [available].
       * By default, this is not allowed ([`Never]). Selecting something
       * different extends the search space enormously.
       *
       * SOURCE EXAMPLE:
       *
       * - [available]: The list of installed packages + The list of
       *   non-installed source packages
       * - [selected]: The list of installed packages + The packages the user
       *   wishes to build
       *
       * BINARY EXAMPLE:
       *
       * - [available]: The list of installed packages + The list of
       *   non-installed binary packages
       * - [selected]: The list of installed packages + The packages the user
       *   wishes to add
       *
       * It is possible to mix source and binary packages. If there are
       * several versions for a package, the most recent version is selected.
       * Installed packages are preferred over binaray packages, which
       * are preferred over source packages.
       *)


val get_required_packages :
      active:package list ->      (* Usually: Which packages are installed *)
      selected:package list ->    (* The selected packages *)
      unit ->
	package list
      (* This is a "soft version" of [analyze_requirements]. It returns the
       * active packages that are required by selected packages. The 
       * resulting list is a subset of the list
       * of active packages only.
       *
       * The function does not check whether the version conditions of
       * dependencies match, nor whether any requirement is missing.
       * Only dependencies, not build dependencies are analyzed.
       * 
       * The active packages should not contain several packages with the
       * same name; otherwise strange results may be returned.
       *
       * Example:
       * - active = all installed packages
       * - selected = one installed package
       * - Result: the installed packages required by the single selected
       *   package
       *)




val get_dependent_packages :
      active:package list ->      (* Usually: Which packages are installed *)
      selected:package list ->    (* The selected packages *)
      unit ->
	package list

      (* Returns the active packages that are dependent on selected 
       * packages. The resulting list is a subset of the list of active
       * packages only. 
       *
       * The function does not check whether the version conditions of
       * dependencies match, nor whether any requirement is missing.
       * Only dependencies, not build dependencies are analyzed.
       * 
       * The active packages should not contain several packages with the
       * same name; otherwise strange results may be returned.
       *
       * Example:
       * - active = all installed packages
       * - selected = selected source packages
       * - Result: The installed packages that are dependent on the
       *   selected source packages - these are the packages that need
       *   to be rebuilt.
       *)

val sort_packages : package list -> package list
      (* Sorts the packages topologically; the deepest ancestor comes
       * first. Duplicates are removed.
       *)

val update_plan : 
      ?replace:replacement_strategy ->
      iteration:int ->
      package list -> 
	int
      (* Plans the build process, and updates the user_status field of the
       * packages.
       *
       * The idea of this function is to feed all package objects as input
       * into it, and let this function decide what to do by updating the
       * user_status fields. What can happen:
       *
       * - Installed packages can be set to `Remove_dep, because they
       *   need to be rebuilt, and this implies the installed package
       *   must be removed first
       * - Installed packages can be set to `Remove_dep because of
       *   dependencies on `Remove packages
       * - Source packages can be set to `Build_dep, because they
       *   need to be rebuilt as result of dependency analysis
       * - Binary packages can be set to `Add_dep, because they
       *   need to be added as result of dependency analysis
       *
       * `Remove_dep, `Build_dep, and `Add_dep have an iteration number as
       * argument; this is set to the passed value of ~iteration.
       *
       * The function returns the number of packages that get `Build_dep,
       * `Remove_dep, or `Add_dep for the first time. If this number is
       * positive, there are packages that still need to be confirmed, and
       * the user dialog to select packages should be shown again (next
       * iteration).
       *
       * @param replace By default, it is not checked whether an already
       * installed package must be replaced by a different version
       * (unless the package is rebuilt anyway). This is [replace=`Never].
       * By passing other values as [replace], this check can be enabled.
       * Warning: This can consume a lot of time!
       *
       * PRECONDITIONS:
       *
       * - It is illegal to have two source or binary packages with the
       *   same name that are both selected for build, or for addition,
       *   respectively. (Exception Package_installation_not_unique.)
       *)
