(* $Id: godi_update.ml 678 2007-02-25 17:16:54Z gerd $
 * ----------------------------------------------------------------------
 * Copyright 2003 by Gerd Stolpmann.
 *
 * This file is part of godi-tools.
 *
 * godi-tools is free software; you can redistribute it and/or modify it
 * under the conditions in the file LICENSE.
 *
 * godi-tools is distributed WITHOUT ANY WARRANTY, as detailed in the
 * file LICENSE.
 *)

open Godi_types
open Godi_exec
open Godi_query
open Godi_versions
open Printf

(* 1. available.new ist eine Temporrdatei -> /tmp/xxxx/available.new
 *    Wenn Download schiefgeht: nicht schlimm, Fehlermeldung ausgeben,
 *    nichts in available_current eintragen.
 * 2. Variable available_current:
 *    - enthlt auch den Prfix des Servers
 *    - Wenn Paket mehrfach da: hhere Versionen schlagen niedrigere.
 *      Sonst nach Reihenfolge vorgehen.
 *)

let available_current = Hashtbl.create 200;;


let words_re = Str.regexp "[ \t\r\n]+";;

let parse_available ?protocol site =
  let out s =
    match protocol with
	None -> prerr_string s
      | Some p -> collect p s
  in

  let bbase = buildbase() in
  let avail_cur = Filename.concat bbase "available.current" in
  let f = 
    try
      open_in avail_cur
    with
	Sys_error msg ->
	  user_error("System error: " ^ msg)
  in
  let l = ref [] in
  let n = ref 0 in
  ( try
      while true do
	incr n;
	let line = input_line f in
	let words = Str.split words_re line in
	match words with
	    [ name; dir; version ] ->
	      l := (name, (dir, version, site)) :: !l
	  | _ ->
	      out (sprintf "Warning: Parsing error in line %d of %s (site %s)\n"
		     !n avail_cur site);
      done;
      assert false
    with
	End_of_file ->
	  close_in f;
      | error ->
	  close_in f;
	  raise error
  );
  (* Now [!l] are the packages from this site. Merge into 
   * available_current: 
   *)
  List.iter
    (fun (name, (dir, version, site)) ->
       let newer =
	 try
	   let (_, version', _) = Hashtbl.find available_current name in
	   let (_, v, rev)   = split_name_version ("x-" ^ version) in
	   let (_, v', rev') = split_name_version ("x-" ^ version') in
	   let d =
	     match cmp_version_string v v' with
		 0 -> rev - rev'
	       | n -> n in
	   d > 0
	 with
	     Not_found -> true in
       if newer then
	 Hashtbl.replace available_current name (dir, version, site)
    )
    !l;
  ()
;;


let update_available ?interactivity ?protocol () =
  let out s =
    match protocol with
	None -> prerr_string s
      | Some p -> collect p s
  in

  Hashtbl.clear available_current;
  let makefile = get_tools_makefile() in

  (* Get [sites], the list of GODI sites with build instructions: *)
  let b_sites = Buffer.create 100 in
  exec_to_user_error
    ~pref_user_error:true
    (call_cmd
       ~stdout:(`Buffer b_sites)
       ?stderr:protocol
       ?interactivity
       ~cd:(buildbase()))
    ("godi_console", [| "make"; "-f"; makefile; "get_sites" |]);
  let sites = Str.split words_re (Buffer.contents b_sites) in

  (* For every site, download "available.current", and merge its
   * contents with the variable available_current
   *)
  List.iter
    (fun site ->
       (* Download available.current for [site]: *)
       exec_to_user_error
         ~pref_user_error:true
         (fun () ->
	    try
	      let args = [| "-f"; makefile;
			    "update_available";
			    ("target_dir=" ^ buildbase() );
			    ("site=" ^ site)
			 |] in
	      call_cmd
		?stdout:protocol
		?stderr:protocol
		?interactivity
		~cd:(buildbase())
		("godi_console", (Array.append [| "make" |] args));
	    with
		Command_failure(_,_,_) ->
		  out ("Error: Cannot retrieve package list from " ^ site ^ " (ignored)\n");
	 )
         ();
       parse_available ?protocol site
	 (* also merges the parsed file into available_current *)
    )
    sites
;;


let install_build_instructions_for ?interactivity ?protocol
                                   site pkgname version pkgrev =

  let makefile = get_tools_makefile() in
  let args =
    [| "-f"; makefile;
       "install_build_instructions";
       "pkgname=" ^ pkgname;
       "version=" ^ version;
	 "pkgrev=" ^ string_of_int pkgrev;
	 "site=" ^ site;
    |] in

  exec_to_user_error
    ~pref_user_error:true
    (call_cmd
       ?stdout:protocol
       ?stderr:protocol
       ?interactivity)
       ("godi_console", (Array.append [| "make" |] args))
;;


let update_build_instructions_for ?interactivity ?protocol pkgname =
  let (dir,version_pkgrev,site) =
    try
      Hashtbl.find available_current pkgname
    with
	Not_found ->
	  user_error ("No such package: " ^ pkgname) in

  let (_, avail_version, avail_pkgrev) =
    split_name_version ("x-" ^ version_pkgrev) in

  (* Now, if [version] is newer than the installed instance, the
   * build.tgz must be downloaded and extracted.
   *)
  let src_pkgs = scan_source_packages ?protocol () in
  let is_newer =
    try
      let inst_pkg = List.find (fun pkg ->
				  (source_pkg pkg)#name = pkgname) src_pkgs in
      let inst_version = (source_pkg inst_pkg) # version in
      let inst_pkgrev  = (source_pkg inst_pkg) # pkgrevision in

      let n1 = cmp_version_string avail_version inst_version in
      let n2 = if n1=0 then avail_pkgrev - inst_pkgrev else n1 in

      n2 > 0

    with
	Not_found -> true in

  if is_newer then
    install_build_instructions_for
      ?interactivity ?protocol site pkgname avail_version avail_pkgrev
;;


let update_build_instructions ?interactivity ?protocol  () =
  Hashtbl.iter
    (fun pkgname _ ->
       update_build_instructions_for ?interactivity ?protocol pkgname)
    available_current
;;
